# DNA Manifest Specification v0.2.1

## Some notes
- The DNA Manifest specification does not include zome references, because those would happen as macros in the development environment.
- This tree is JSON only as convenience, it could be s-expressions or any other tree structure.
- Two levels: the DNA has a manifest level which gets interpreted by the Holochain Core library, and the WASM level which gets interpreted by the Ribosome.
It's not allways clear what goes in which level. For example, schema used to live at the Holochain Core library level under the `schema` property, because spinning the Ribosome to do frequent schema matching didn't seem to make sense.
- The manifest is to be generated by a packaging tool 'hc' / 'DNA Builder', which figures it out with just the zome code (zome code has macros for outputting to JSON according to this specification).

## File naming

- DNA manifest file name is: `<name>.dna.json`
- DNA-template manifest file name is: `<name>.dna-tpl.json`

## Design
Function declarations (i.e. names and signatures) are shown outside of WASM in the manifest, kind of like C header files, because they are used for validating function calls according to capability membranes at the Holochain Core library level, not the WASM level.

### WASM
One WASM module per zome: All different Capability zome code and validation zome code is to be be compiled into one WASM module.

### No more links as entry-types

We want to remove the notion of links as an entry format.
Instead, declarations about linking are in the entry-type definition and use that information to dispatch validation, instead of dispatching off of the linked entries type names.

For example:

``` javascript
// OLD WAY
// Note that you write validation by checking the type in the call to
// `validateLink`
commit("post",{content:"my post"})
commit("post_link",[
  {base:post,tag:"handle",target:myHandle},
  {base:myHandle,tag:"posts",target:post}
])
commit("tag_link",[
  {base:myHandle,tag:"_tag"+theTag, target:post}
  {base:post,tag:"taggedBy",target:myHandle},
])

// NEW WAY
// Note: The links' validation code is in the entry-type's `links-to` and `links-from` block
// which gets found by lookup via the type of the base and the tag.
hc_start_bundle_mode();
{
  let post = hc_commit_entry("post",{content:"my post"});
  hc_commit_link(post, "author", myHandle);
  hc_commit_link(myHandle, "posts", post);
  hc_commit_link(myHandle, "_tag" + theTag, post);
  hc_commit_link(post,"taggedBy", myHandle);
}
hc_end_bundle_mode(FINISH_BUNDLE); // or CANCEL_BUNDLE
```

## Specification as example DNA Manifest

``` javascript
{
  // Config: required settings for all DNAs used internally by Holochain Core.
  // Available in zome code as global constants.
  "manifest_spec_version": "2.0",
  "dna_name": "SomeName",
  "description": "this app does very cool stuff",
  "version": "0.0.1",
  "uuid": "123e4567-e89b-12d3-a456-426655440000",
  "dht_config": {
    "HashType": "sha2-256"
    // ...
  }
  // ...
  // Properties: Settings specific to a dna-template, used only in zome code.
  // Available in zome code with the `hc_property()` Zome API Function.
  // Note: Later we might also want have properties defined within a Zome, and possibly move this part.
  "properties": {
    "language": "en"
  },
  // Code: Set of zomes
  "zomes": [
    {
      // Zome config
      "zome_name": "clutter",
      "description": "zome that implements micro-blogging",
      // Ribosome or HDK level config that might need to be set
      // to indicate how to handle this Zome. Not used yet, but anticipating.
      "config": {},
      // Zome code
      "code": "..", // an s-expression encoded wasm or an Base64 encoded wasm bytecode
      // Zome entry types
      "entry_types": [
        {
          "entry_type_name": "post",
          "description": "A blog post entry which has an author.",
          "sharing": "public",  // or private, encrypted
          "links_to": [
            {
              "target_type": "handle",
              "tag": "author",
            },
          ]
        },
        {
          "entry_type_name": "handle",
          "description": "An agent's handle on the network. The DNA indexes all the handles.",
          "sharing": "public",
          "links_to": [
            {
              "target_type": "post",
              "tag": "posts",
            },
          ]
          "linked_from": [
            {
              "base_type": "__AGENT_HASH__",
              "tag": "agent"
            },
            {
              "base_type": "__DNA_ADDRESS__",
              "tag": "handles"
            },
          ]
        }
      ], // end of entry-types
      // Zome fn_declarations
      "fn_declarations": [
          {
              "name": "newPost",
              "inputs": [
                  {
                      "name": "post",
                      "type": "string"
                  }
              ],
              "outputs": [
                  {
                      "name": "hash",
                      "type": "string"
                  }
              ]
        },
        // ...
      ],
      // Zome traits which group functions for composiblity
      "traits": {
        "hc_public":
        {
          "functions": ["getPost"]
        }
      },
      // Zome signals
      "signals": [
            {
              "sig_name": "posted",
              "description": "signal emmited when a post is committed",
              "config-params": null,
              "arguments": {
                  "name": "hash",
                  "type": "hash"
              },
            // ...
          ],
        },
        "indexing":
        {
          "type": "public",
          "functions": [
            // ...
          ],
        },
      }
    }
  ] // end of zomes
}
```
